package body Contenedores.rboles_Binarios.Ajustables.Biselados is
	function Raz(C: in Biselado) return TElemento is
	begin
		return C.Elementos.Clave;
	end Raz;

	--------------------------
	-- Operaciones Internas --
	--------------------------

	-- Biselar realiza un paso de la biselacin, ascendiendo un nodo
    -- mediante una rotacin
	procedure Biselar(
			Raz, A: in out rbol_Bin_bus;
			Haca  : Orientacin;
			E      : TElemento;
			Dos_Niveles: in out Integer
			) is
		Rama : Integer := Orientacin'Pos(haca) + 1;
		Rama2: Integer;
		Hijo : Rbol_bin_Bus;
	begin
		if A = Raz and Dos_niveles > 0 then
			-- En el nivel de la raz se realiza un rotacin simple
			-- si no existe abuelo
			if A.Hijos(Rama) /= null then
				Rotar_Simple(A, Haca);
			end if;
		else
			if Dos_niveles > 0 then
				Dos_Niveles := Dos_Niveles - 1;
			else
				-- Asciende otro nivel, existe abuelo (Dos_Niveles = 0)
				Hijo := A. Hijos(Rama);
				if E < Hijo.Clave then
					Rama2 := 1;
				else
					Rama2 := 2;
				end if;
				if Rama = Rama2 then
					Rotar_Doble_Inversora(A, Haca);
				else
					Rotar_Doble(A, Haca);
				end if;
				Dos_Niveles := 1;
			end if;
		end if;
	end Biselar;

	-- Biselar_subrbol bisela el nodo con el menor o el mayor
	-- valor de clave de un rbol, dependiendo del valor de Haca
	procedure Biselar_Subrbol(
			Raz, A: in out rbol_Bin_bus;
			Haca  : Orientacin;
			Dos_Niveles: in out Integer
			) is
		Rama : Integer := Orientacin'Pos(haca) + 1;
	begin
		if A = null then
			Dos_niveles := 2;
		else
			-- Descenso por la misma rama
			Biselar_Subrbol(Raz, A.Hijos(Rama), Haca, Dos_Niveles);
			if A = Raz and Dos_niveles > 0 then
				if A.Hijos(Rama) /= Null then
					Rotar_Simple(A, Haca);
				end if;
			else
				if Dos_Niveles > 0 then
					Dos_niveles := Dos_Niveles - 1;
				else
					-- Asciende otro nivel,
					-- existe abuelo (Dos_Niveles = 0)
					-- Se aplica una rotacin doble inversora porque
 				    -- se ha descendido siempre por la misma rama
					Rotar_Doble_Inversora(A, Haca);
					Dos_Niveles := 1;
				end if;
			end if;
		end if;
	end Biselar_Subrbol;

	procedure Insertar(
			Raz, A: in out rbol_Bin_Bus;
			E	   : in TElemento; 
			Dos_Niveles: in  out Integer
			) is
		Rama: orientacin;
	begin
		if A = null then
			-- El nuevo valor se inserta como hoja
			A := new Nodo_binario'(E, (null, null));
			Dos_Niveles := 1;
		else
			-- Si E = A.Clave, no se inserta
			if E = A.Clave then
				Dos_Niveles := 1;
			else
				-- La insercin continua en el subrbol adecuado
				if E < A.Clave then
					Insertar(Raz, A.Hijos(1), E, Dos_Niveles);
					Rama := Izquierda;
				else
					if A.Clave < E then
						Rama := Derecha;
						Insertar(Raz, A.Hijos(2), E, Dos_Niveles);					
					end if;
				end if;
				-- El ltimo nodo en que se estuvo se asciende u paso
				biselar(Raz, A, Rama, E, Dos_niveles);
			end if;
		end if;
	end Insertar;

	procedure Buscar(
			Raz, A: in out rbol_Bin_Bus;
			E      : in TElemento; 
			Dos_Niveles: in out Integer;
			Encontrado : in out Boolean
			) is
		Rama : Orientacin;
	begin
		if A = null then
			-- La clave buscada no est en el rbol
			Dos_Niveles := 2;
			Encontrado := False;
		else
			if E = A.Clave then
				-- Se ha encontrado
				Dos_Niveles := 1;
				Encontrado := True;
			else
				-- La bsqueda contina en el subrbol adecuado
				if E < A.Clave then
					Buscar(Raz, A.Hijos(1), E, Dos_niveles, Encontrado);
					Rama := izquierda;
				else
					Buscar(Raz, A.Hijos(2),E, Dos_niveles, Encontrado);
					Rama := Derecha;
				end if;
				-- El ltimo nodo en que se estuvo se asciende un paso
				Biselar(Raz, A, Rama, E, Dos_Niveles);
			end if;
		end if;
	end Buscar;

	-------------------------------
	-- Operaciones constructoras --
	-------------------------------	
	procedure Insertar(C: in out Biselado; E: in TElemento) is
		Dos_Niveles: Integer := 0;
	begin
		Insertar(C.Elementos, C.Elementos, E, Dos_Niveles);
	end;

	procedure Extraer(C: in out Biselado;E: in TElemento) is
		Dos_Niveles : Integer := 0;
		Encontrado  : Boolean := False;
		Izqdo, Drcho: Rbol_Bin_bus;
	begin
		-- Buscar bisela hasta la raz el nodo que contiene la clave 
		-- Buscada, o el ltimo accedido en el camino de bsqueda
		Buscar(C.Elementos, C.Elementos, E, Dos_Niveles, Encontrado);
		if Encontrado then
			-- Si se encontr, se biselo a la raz. Nos quedamos con los
			-- hijos de sta y la eliminamos
			Izqdo := C.Elementos.Hijos(1);
			Drcho := C.Elementos.Hijos(2);
			Eliminar(C.Elementos);

			-- Se bisela el sucesor o el predecesor en simtrico
			-- y se enganchan los dos subrboles
			if Drcho /= null then
				C.Elementos := Drcho;
				Biselar_subrbol(C.Elementos, C.Elementos, Izquierda, Dos_niveles);
				C.Elementos.Hijos(1) := Izqdo;
			else
				C.Elementos := Izqdo;
				Biselar_subrbol(C.Elementos, C.Elementos, Derecha, Dos_niveles);
			end if;
		end if;
	end;

	procedure Vaciar(C: in out Biselado) is
	begin
		Vaciar(Binario_bsqueda(C));
	end;

	------------------------------
	-- Operaciones observadoras --
	------------------------------
	procedure Est(C: in out Biselado; E: in TElemento; Encontrado: out Boolean) is
		Dos_Niveles: Integer := 0;
	begin
		Encontrado := False;
		Buscar(C.Elementos, C.Elementos, E, Dos_niveles, Encontrado);
	end;

	function Tamao(C: in Biselado) return Natural is
	begin
		return Tamao(Binario_Bsqueda(C));
	end;
end Contenedores.rboles_Binarios.Ajustables.Biselados;
