#ifndef INC_BISELADO
#define INC_BISELADO
#include "autoajustable.h"
#include "nodobinario.h"

template <class T>
class Biselado:public AutoAjustable<T,NodoBinario<T> > {
      NodoBinario<T> *raiz;
      bool buscarInterno(NodoBinario<T> *&,const T&,int &);
   	/**
   	 * Aade el elemento valor al contenedor, si no est ya presente.
   	 * En cualquier caso, lo bisela a la raz.
   	 */
      void insertarInterno(NodoBinario<T>*&,const T&,int &);
      void pasoBiselacion(NodoBinario<T>*&,int,const T&,int &);    
      void biselaSubarbol(NodoBinario<T>*&,int,int &); 
   public:
      Biselado(){
         raiz=NULL;}
      ~Biselado();
      void insertar(const T&);
      void extraer(const T&);
      bool encuentra(const T&)const;
      bool encuentra(const T&);
      void vaciar();
};

// Funciones miembro de la clase Biselado
template <class T>
   Biselado<T>::~Biselado()
   {
      eliminarInterno(raiz);
   }

template <class T>
   bool Biselado<T>::buscarInterno(NodoBinario<T> *&nodo,const T& valor,int &dosNiveles)
   {
      int rama;
      bool encontrado;
      if(nodo==NULL)
      {
         dosNiveles=2;
         return false;
      }
      else
         if(valor==nodo->clave)
         {
            dosNiveles=1;
            return true;
         }
         else
         {
            if(valor<nodo->clave)
               rama=0;
            else
               rama=1;
            encontrado=buscarInterno(nodo->enlaces[rama],valor,dosNiveles);
            pasoBiselacion(nodo,rama,valor,dosNiveles);
            return encontrado;
         }
   }

template <class T>
   void Biselado<T>::insertarInterno(NodoBinario<T>*&nodo,const T& valor,int &dosNiveles)
   {
      int rama;
      if(nodo==NULL)
      {
        // El nuevo valor se inserta como hoja
         nodo=new NodoBinario<T>(valor);
         this->numElem++;
         dosNiveles=1;
      }
      else
         if(valor==nodo->clave)
         {
            dosNiveles=1;
         }
         else
         {
            if(valor<nodo->clave)
               rama=0;
            else
               rama=1;
            insertarInterno(nodo->enlaces[rama],valor,dosNiveles);
            pasoBiselacion(nodo,rama,valor,dosNiveles);
         } 
   }

template <class T>
   void Biselado<T>::pasoBiselacion(NodoBinario<T>*&nodo,int rama,const T& valor,int &dosNiveles)
   {
      int ramaDos;
      NodoBinario<T> *hijo;
      if(nodo==raiz && dosNiveles>0) //nivel de la raz y no existe abuelo
      {
         if(nodo->enlaces[rama]!=NULL)
            rotarSimple(nodo,rama);
      }
      else
         if(dosNiveles>0)
            dosNiveles--;                     //asciende otro nivel
         else                                //existe abuelo (dosNiveles=0) 
         {
            hijo=nodo->enlaces[rama];
            if(valor<hijo->clave)
               ramaDos=0;
            else
               ramaDos=1;
            if(rama==ramaDos)
               rotarDobleInversora(nodo,rama);
            else
               rotarDoble(nodo,rama);
            dosNiveles=1;
         }
   }    

template <class T>
   void Biselado<T>::biselaSubarbol(NodoBinario<T>*&nodo,int rama,int &dosNiveles)
   {
      if(nodo==NULL)
         dosNiveles=2;
      else
      {
      //descenso por la misma rama
         biselaSubarbol(nodo->enlaces[rama],rama,dosNiveles);
         if(nodo==raiz && dosNiveles>0)
         {
            if(nodo->enlaces[rama]!=NULL)
               rotarSimple(nodo,rama);
         }
         else
            if(dosNiveles>0)
               dosNiveles--;                     //asciende otro nivel
            else                                //existe abuelo (dosNiveles=0) 
            {
            //se aplica una rotacin doble inversora porque se ha descendido siempre por la misma rama
               rotarDobleInversora(nodo,rama);
               dosNiveles=1;
            }
      }
   }    

template <class T>
   void Biselado<T>::insertar(const T& valor)
   {
      int dosNiveles=0;
      insertarInterno(raiz,valor,dosNiveles);
   }

template <class T>
   void Biselado<T>::extraer(const T& valor)
   {
      int dosNiveles;
      NodoBinario<T> *hijo, *otroHijo;
      bool encontrado=buscarInterno(raiz,valor,dosNiveles);
      if(encontrado)
      {
         hijo=raiz->enlaces[0];
         otroHijo=raiz->enlaces[1];
         delete raiz;
         this->numElem--;
         if(otroHijo!=NULL) //se bisela el sucesor en simtrico del valor de clave extrado
         {
            raiz=otroHijo;
            biselaSubarbol(raiz,0,dosNiveles);
            raiz->enlaces[0]=hijo;
         }
         else  //no existe el sucesor, se bisela el predecesor 
         {
            raiz=hijo;
            biselaSubarbol(raiz,1,dosNiveles);
         }
      }
   }

template <class T>
   bool Biselado<T>::encuentra(const T& valor)const
   {
      return BinarioBusquedaBasico<T,NodoBinario<T> >::buscarInterno(raiz,valor);
   }

template <class T>
   bool Biselado<T>::encuentra(const T& valor)
   {
      int dosNiveles=0;
      return buscarInterno(raiz,valor,dosNiveles);
   }

template <class T>
   void Biselado<T>::vaciar()
   {
      eliminarInterno(raiz);
      raiz=NULL;
      this->numElem=0;
   }

#endif



