package rbolesBinariosBsqueda;

import contenedores.ContenedorOrdenado;

/**
 * Implementacin de un contenedor usando una estructura de rbol biselado
 */
public class rbolBiselado
	extends rbolBinarioAjustable
	implements ContenedorOrdenado {

	/**
	 * clase interna auxiliar
	 */
	private class EnteroMutable {
		int entero;
		public EnteroMutable() {
			entero = 0;
		}
		public void incrementar(){
			entero++;
		}
		public void decrementar(){
			entero--;
		}
		public int valor(){
			return entero;
		}
		public void ponerValor(int n){
			entero=n;
		}
	}

	private NodoBinario raz;

	public rbolBiselado() {
		super();
		raz = null;
	}

	/**
	 * Realiza un paso de la biselacin, ascendiendo un nodo 
	 * mediante una rotacin
	 */
	private NodoBinario bisela(
		NodoBinario nodo,
		int rama,
		Comparable valor,
		EnteroMutable niveles) {
		int ramaDos;
		NodoBinario hijo;

		if ((nodo == raz) && (niveles.valor() > 0)) {
			/*
			* En el nivel de la raz se realiza una rotacin simple si
			* no existe abuelo
			*/
			if (nodo.enlaces[rama] != null) {
				nodo = rotacinSimple(nodo, rama);
			}
		} else {
			if (niveles.valor() > 0) {
				// An no se ha llegado al abuelo
				niveles.decrementar();
			} else {
				// El nodo actual asciende dos niveles 
				hijo = nodo.enlaces[rama];
				if (valor.compareTo(hijo.clave) < 0)
					ramaDos = 0;
				else
					ramaDos = 1;
				if (rama == ramaDos)
					nodo = rotacinDobleInversora(nodo, rama);
				else
					nodo = rotacinDoble(nodo, rama);
				niveles.ponerValor(1);
			}
		}
		return nodo;
	}

	/**
	 * Aade el elemento valor al contenedor, si no est ya presente.
	 * En cualquier caso, lo bisela a la raz.
	 */
	private NodoBinario insertar(
		NodoBinario nodo,
		Comparable valor,
		EnteroMutable niveles) {
		int rama;
		if (nodo == null) {
			// El nuevo valor se inserta como hoja
			nodo = new NodoBinario(valor);
			numElem++;
			niveles.ponerValor(1);
		} else {
			// si valor es igual a nodo.clave no se inserta
			if (valor.compareTo(nodo.clave) == 0) {
				niveles.ponerValor(1);
			} else {
				// La insercin contina en el subrbol adecuado
				if (valor.compareTo(nodo.clave) < 0)
					rama = 0;
				else
					rama = 1;
				nodo.enlaces[rama] =
					insertar(nodo.enlaces[rama], valor, niveles);

				// El nodo actual se bisela un paso
				nodo = bisela(nodo, rama, valor, niveles);
			}
		}
		return nodo;
	}

	/* (non-Javadoc)
	 * @see contenedores.ContenedorOrdenado#insertar(java.lang.Comparable)
	 */
	public void insertar(Comparable valor) {
		EnteroMutable niveles = new EnteroMutable();
		raz = insertar(raz, valor, niveles);
	}

	/* (non-Javadoc)
	 * @see contenedores.ContenedorOrdenado#extraer(java.lang.Comparable)
	 * Se quita el elemento valor del contendor, transformando en cualquier
	 * caso el rbol segn las especificaciones de un rbol biselado.
	 */
	 public void extraer(Comparable valor) {
		 EnteroMutable niveles = new EnteroMutable();
		 NodoBinario hijo, otroHijo;

		 /*
		  * est bisela hasta la raz el nodo que contiene la clave buscada,
		  * o el ltimo accedido en el camino de bsqueda.
		  */
		 raz = est(raz, valor, niveles);
		 if ((raz != null) && (valor.compareTo(raz.clave) == 0)) {
			 /*
			  * Si se encontr, se bisel a la raz. Nos quedamos con los
			  * hijos de sta y la eliminamos
			  */
			 hijo = raz.enlaces[0];
			 otroHijo = raz.enlaces[1];
			 numElem--;

			 /*
			  * Se bisela el sucesor o el predecesor en simtrico y se
			  * enganchan los dos subrboles
			  */
			 if (otroHijo != null) {
				 raz = otroHijo;
				 raz = biselaSubrbol(otroHijo, 0, niveles);
				 raz.enlaces[0] = hijo;
			 } else{
				 raz = hijo;
				 raz = biselaSubrbol(hijo, 1, niveles);
			 }
		 }
	 }
	
	 /**
	  * Bisela el nodo con el mayor o el menor valor de clave de un rbol
	  * dependiendo del valor de rama
	  */
	 private NodoBinario biselaSubrbol(
		 NodoBinario nodo,
		 int rama,
		 EnteroMutable niveles) {

		 if (nodo == null)
			niveles.ponerValor(2);
		 else {
			 // Descenso por la misma rama
			 nodo.enlaces[rama] =
				 biselaSubrbol(nodo.enlaces[rama], rama, niveles);
			 if (nodo == raz && niveles.valor() > 0) {
				 if (nodo.enlaces[rama] != null)
					 nodo = rotacinSimple(nodo, rama);
			 } else if (niveles.valor() > 0) {
				 niveles.decrementar();
			 } else {
				 /*
				  * Se aplica una rotacin doble inversora porque se ha
				  * descendido siempre por la misma rama
				  */
				 nodo = rotacinDobleInversora(nodo, rama);
				 niveles.ponerValor(1);
			 }
		 }
		 return nodo;
	 }

	/* (non-Javadoc)
	 * @see contenedores.ContenedorOrdenado#est(java.lang.Comparable)
	 */
	public boolean est(Comparable valor) {
		EnteroMutable niveles = new EnteroMutable();
		raz = est(raz, valor, niveles);
		if ((raz != null) && (valor.compareTo(raz.clave) == 0))
			return true;
		else
			return false;
	}

	/**
	 * Devuelve el nodo que bisela a la raz del rbol (el ltimo nodo
	 * visitado en el camino de bsqueda)
	 */
	private NodoBinario est(
		NodoBinario nodo,
		Comparable valor,
		EnteroMutable niveles) {
		int rama;
		if (nodo == null) {
			// La clave buscada no est en el rbol
			niveles.ponerValor(2);
		} else {
			if (valor.compareTo(nodo.clave) == 0) {
				// Se ha encontrado
				niveles.ponerValor(1);
			} else {
				// La bsqueda contina en el subrbol adecuado
				if (valor.compareTo(nodo.clave) < 0)
					rama = 0;
				else
					rama = 1;
				nodo.enlaces[rama] =
					est(nodo.enlaces[rama], valor, niveles);

				// El nodo actual se bisela un paso
				nodo = bisela(nodo, rama, valor, niveles);
			}
		}
		return nodo;
	}
	
	/* (no Javadoc)
	 * @see contenedores.ContenedorOrdenado#vaciar()
	 */
	public void vaciar() {
		raz = null;
		numElem = 0;
	}
}
